// by Alexander Ilyin
// make access to files to be more convenient

#include "file.h"
#include "error.h"
#include <string>
using namespace std;


// *for convenience
LARGE_INTEGER offset_0 = {0};
// *temporary transient macro
#define SA A


// need these 64 bit functions; MSVC 6.0 libraries don't have them
#if _MSC_VER <= 1200
typedef BOOL WINAPI GetFileSizeEx_t (HANDLE hfile, PLARGE_INTEGER file_size);
typedef BOOL WINAPI SetFilePointerEx_t (HANDLE hfile, LARGE_INTEGER distance_to_move, LARGE_INTEGER* new_file_pointer, DWORD move_method);
GetFileSizeEx_t* GetFileSizeEx = 0;
SetFilePointerEx_t* SetFilePointerEx = 0;

static bool InitFunctions ()
{
	HMODULE h_kernel32;

	h_kernel32 = GetModuleHandle (TEXT("KERNEL32"));
	A(h_kernel32);
	*(FARPROC*)&GetFileSizeEx = GetProcAddress (h_kernel32, "GetFileSizeEx");
	if (! GetFileSizeEx)
		return 0;
	*(FARPROC*)&SetFilePointerEx = GetProcAddress (h_kernel32, "SetFilePointerEx");
	if (! SetFilePointerEx)
		return 0;
	return 1;
}
#endif

void File::GetError ()
{
	DWORD D;

	D = GetLastError ();
	switch (D)
	{
	case ERROR_INVALID_HANDLE:
// *this should be some our internal error
		ET;
		break;
	case ERROR_FILE_NOT_FOUND:
		m_error = FILE_ERR__NO_FILE;
		break;
	case ERROR_SHARING_VIOLATION:
	case ERROR_ACCESS_DENIED:
		m_error = FILE_ERR__NO_ACCESS;
	case ERROR_HANDLE_DISK_FULL:
	case ERROR_DISK_FULL:
		m_error = FILE_ERR__DISK_FULL;
		break;
	}
}

bool File::Open (bool write, bool widechar, const void* filepath)
{
//	BOOL B_r;
	DWORD access;
	DWORD share;
	DWORD disposition;

// not very nice place for that, but it isn't so horrible
#if _MSC_VER <= 1200
	bool r;
	r = InitFunctions ();
	if (! r)
		return 0;
#endif

#if 0
// *members are not zeroed by the default constructor
	memset (this, 0, sizeof(*this));
#endif
	A(! m_is_opened);
	A(! m_hfile);
	m_error = FILE_ERR__UNSPECIFIED_OPEN_ERROR;
	m_write = write;

#if 0
// impossible, because it was just zeroed a moment ago ^^
// commander sequence error
	if (m_hfile)
		return 0;
#endif
#if 0
// save the filepath for a possible usage in the future
//	memcpy (m_arc_name, filepath, strlen ((char*)filepath) + 1);
	strncpy (m_arc_name, (char*)filepath, sizeof(m_arc_name) - 1);
	m_arc_name[sizeof(m_arc_name) - 1] = 0;
#endif
	access = write ? GENERIC_WRITE : GENERIC_READ;
// *"add to archive" is never useful; always need to have a new, clean archive
// but maybe we have to overwrite an existing file; so have to share it for read
	share = write ? FILE_SHARE_READ : FILE_SHARE_READ | FILE_SHARE_WRITE;
//	share = write ? 0 : FILE_SHARE_READ | FILE_SHARE_WRITE;
	disposition = write ? CREATE_ALWAYS : OPEN_EXISTING;
	if (widechar)
// *it also works without prepending "\\?\" at the path, but then the length will be limited to MAX_PATH
	{
		wstring ws;
		ws = L"\\\\?\\";
		ws += (wchar_t*)filepath;
		m_hfile = CreateFileW (ws.c_str(), access, share, 0, disposition, FILE_ATTRIBUTE_NORMAL, 0);
	}
	else
		m_hfile = CreateFileA ((PCSTR)filepath, access, share, 0, disposition, FILE_ATTRIBUTE_NORMAL, 0);
	if (m_hfile == INVALID_HANDLE_VALUE)
//	if (m_hfile == -1)
	{
		GetError ();
		return 0;
	}
	m_is_opened = 1;
	m_error = FILE_ERR__NONE;
	return 1;
}

bool File::Close ()
{
	BOOL B_r;

	A(m_is_opened);
	A(m_hfile);
	m_error = FILE_ERR__CLOSE_ERROR;
//	if (! m_hfile)
//		return 0;
	B_r = CloseHandle (m_hfile);
	m_hfile = 0;
	m_is_opened = 0;
	if (! B_r)
		return 0;
	m_error = FILE_ERR__NONE;
	return 1;
}

bool File::GetSize ()
{
	BOOL B_r;

//	A(m_is_opened);
	A(m_hfile);
	A(! m_write);
	m_error = FILE_ERR__READ_FAULT;
	SA(sizeof(m_len) == sizeof(LARGE_INTEGER));
	B_r = GetFileSizeEx (m_hfile, (LARGE_INTEGER*)&m_len);
	if (! B_r)
	{
		GetError ();
		return 0;
	}
	m_error = FILE_ERR__NONE;
	return 1;
}

bool File::GetTime (FILETIME& ft)
{
	BOOL B_r;

//	A(m_is_opened);
	A(m_hfile);
	A(! m_write);
	m_error = FILE_ERR__READ_FAULT;
	B_r = GetFileTime (m_hfile, 0, 0, &ft);
	if (! B_r)
	{
		GetError ();
		return 0;
	}
	m_error = FILE_ERR__NONE;
	return 1;
}

bool File::SetTime (FILETIME& ft)
{
	BOOL B_r;

//	A(m_is_opened);
	A(m_hfile);
	A(m_write);
	m_error = FILE_ERR__WRITE_FAULT;
	B_r = SetFileTime (m_hfile, 0, 0, &ft);
	if (! B_r)
	{
		GetError ();
		return 0;
	}
	m_error = FILE_ERR__NONE;
	return 1;
}

bool File::IsNotEOF (bool& is_not_eof)
{
	BOOL B_r;
	LARGE_INTEGER pos;
	LARGE_INTEGER len;

//	A(m_is_opened);
	A(m_hfile);
// *this operation has no sense for write
	A(! m_write);
	is_not_eof = 0;
	m_error = FILE_ERR__READ_FAULT;
	if (! m_hfile)
		return 0;
// get the current position of file pointer
	B_r = SetFilePointerEx (m_hfile, offset_0, &pos, FILE_CURRENT);
	if (! B_r)
		goto error;
	B_r = GetFileSizeEx (m_hfile, &len);
	if (! B_r)
		goto error;
//	if (pos > len)
	if (pos.QuadPart > len.QuadPart)
		return 0;
	is_not_eof = pos.QuadPart < len.QuadPart;
	m_error = FILE_ERR__NONE;
	return 1;

error:
	GetError ();
	return 0;
}

bool File::SetPointer (bool from_current, u64 pos)
{
	BOOL B_r;
	LARGE_INTEGER pos0;
	LARGE_INTEGER pos1;
	LARGE_INTEGER i;
	DWORD dispos;

	m_error = FILE_ERR__READ_FAULT;

//	A(m_is_opened);
	A(m_hfile);
// *this operation has no sense for write
	A(! m_write);
//	if (! m_hfile)
//		return 0;
	dispos = from_current ? FILE_CURRENT : FILE_BEGIN;
	if (from_current)
	{
// get the current position
		B_r = SetFilePointerEx (m_hfile, offset_0, &pos0, FILE_CURRENT);
		if (! B_r)
			return 0;
	}
	i.QuadPart = pos;
	B_r = SetFilePointerEx (m_hfile, i, &pos1, dispos);
	if (! B_r)
		return 0;
	if (from_current)
// relative
	{
		if (pos1.QuadPart != pos0.QuadPart + pos)
			return 0;
	}
	else
// absolute
	{
		if (pos1.QuadPart != pos)
			return 0;
	}
	m_error = FILE_ERR__NONE;
	return 1;
}

bool File::Read (void* buf, u32 len)
{
	BOOL B_r;
	DWORD num_read;

//	A(m_is_opened);
	A(m_hfile);
	A(! m_write);
	m_error = FILE_ERR__READ_FAULT;
//	if (! m_hfile)
//		return 0;
	B_r = ReadFile (m_hfile, buf, len, &num_read, 0);
// *is 1, even if the file is empty, or it is EOF ('num_read' then is 0)
	if (! B_r)
		goto error;
	if (num_read != len)
		goto error;
	m_error = 0;
	return 1;

error:
	GetError ();
	return 0;
}

bool File::Write (void* buf, u32 len)
{
	BOOL B_r;
	DWORD num_written;

//	A(m_is_opened);
	A(m_hfile);
	A(m_write);
	m_error = FILE_ERR__WRITE_FAULT;
//	if (! m_hfile)
//		return 0;
	B_r = WriteFile (m_hfile, buf, len, &num_written, 0);
	if (! B_r)
		goto error;
	if (num_written != len)
		goto error;
	m_error = 0;
	return 1;

error:
	GetError ();
	return 0;
}

